# coding: utf-8
'''
Created on 07.08.2015

@author: v.popov
'''
import StringIO
import base64
import datetime
import json
import os
import uuid
from xml.sax import make_parser, ContentHandler

from common.sysfunctions import tableCursorImport
from ru.curs.celesta import CallContext, ConnectionPool, SessionContext
from ru.curs.celesta.showcase.utils import XMLJSONConverter
import re
import xml.etree.ElementTree as ET
from java.lang import String
from java.io import ByteArrayInputStream, ByteArrayOutputStream
from common.dbutils import DataBaseXMLExchange


try:
    from ru.curs.showcase.core.jython import JythonDTO
except:
    from ru.curs.celesta.showcase import JythonDTO
    
try:
    from suds.client import Client
except:
    pass


class parserSAXHandlerSelect(ContentHandler):
    def __init__(self):
        self.pin = None
        self.grain = None
        self.table = None
        self.param = None
        self.sync = 0
        self.systemId = None
        self.recFlag = False
        self.dataWriteFlag = False
        self.commandFlag = False
        self.recordsFlag = False
        self.columnFlag = False
        self.currentNode = None
        self.currentData = None
        self.username = None
        self.columnsToSelect = []

    def startElement(self, name, attrs):
        if name == "command":
            self.commandFlag = True
            for attr, attrValue in attrs.items():
                if attr == "param":
                    self.param = attrValue
                elif attr == "pin":
                    self.pin = attrValue
                elif attr == "grain":
                    self.grain = attrValue
                elif attr == "table":
                    self.table = attrValue
                elif attr == "sync":
                    self.sync = attrValue
                elif attr == "system_id":
                    self.systemId = attrValue
        if not (self.pin or self.grain or self.table) and name in ("pin", "grain", "table"):
            self.dataWriteFlag = True
        if name == "column":
            self.columnFlag = True

    def characters(self, data):
        if self.dataWriteFlag == True:
            self.currentData = data

    def endElement(self, name):
        if name == "pin" and self.dataWriteFlag == True:
            self.pin = self.currentData
            self.dataWriteFlag = False
        elif name == "grain" and self.dataWriteFlag == True:
            self.grain = self.currentData
            self.dataWriteFlag = False
        elif name == "table" and self.dataWriteFlag == True:
            self.table = self.currentData
            self.dataWriteFlag = False
        elif name == "column":
            self.columnsToSelect.append(self.currentData)
            self.columnFlag = False

def select(self, xmlIn):
    if not os.path.isdir(r'C:\accred_service_logs'):
        os.makedirs(r'C:\accred_service_logs')
    dateTime = str(datetime.datetime.now().strftime("%Y-%m-%d %H.%M.%S"))
    mypath = r'C:\accred_service_logs/' + "select_" + dateTime
    if not os.path.isdir(mypath):
        os.makedirs(mypath)
    logFile = open(mypath + "/input.xml", "w")
    logFile.write(xmlIn)
    logFile.close()
    parser = make_parser()
    handler = parserSAXHandlerSelect()
    parser.setContentHandler(handler)
    parser.parse(StringIO.StringIO(xmlIn.encode('utf8')))
    conn = ConnectionPool.get()
    sesContext = SessionContext('master', 'master')
    context = CallContext(conn, sesContext)
    webServiceUsers = tableCursorImport('webservices', 'web_service_users')(context)
    webServiceUsers.get(handler.pin)
    username = getattr(webServiceUsers, 'login')
    sesContext = SessionContext(username, username)
    userContext = CallContext(conn, sesContext)
    currentTable = tableCursorImport(handler.grain, handler.table)(userContext)
    whereClause = handler.param
    if whereClause is not None and whereClause != '':
        currentTable.setComplexFilter(whereClause)
    systemId = handler.systemId
    sync = int(handler.sync)
    if sync == 1 and systemId is not None:
        currentTableMeta = currentTable.meta()
        tableJson = json.loads(currentTableMeta.getCelestaDoc())
        syncGrain, syncTable = tableJson["sync_table"].split('.')
        syncTableCursor = tableCursorImport(syncGrain, syncTable)(context)
        syncPimaryKey = syncTableCursor.meta().getPrimaryKey()
        packageId = str(uuid.uuid4())
        data = {"records":{"@package_id":packageId, "@system_id":systemId, "rec":[]}}
    else:
        data = {"records":{"rec":[]}}
    if currentTable.count() != 0:
        columns = currentTable.meta().getColumns()
        columnsToSelectExist = []
        columnsToSelect = handler.columnsToSelect
        if not columnsToSelect:
            columnsToSelect = columns
        else:
            for column in columnsToSelect:
                if column in columns:
                    columnsToSelectExist.append(column)
        for rec in currentTable.iterate():
            rec_dict = {}
            for column in columnsToSelect:
                rec_dict[column] = getattr(rec, column)
            if sync == 1:
                keyValueDict = {}
                keyValueList = []
                for key in syncPimaryKey:
                    if key in currentTable.meta().getColumns() and key != "system_id":
                        setattr(syncTableCursor, key, getattr(rec, key))
                        keyColumnValue = getattr(rec, key)
                        keyValueDict[key] = keyColumnValue
                        keyValueList.append(keyColumnValue)
                keyValueDict["system_id"] = (systemId)
                keyValueList.append(systemId)
                syncTableCursor.system_id = systemId
                syncTableCursor.package_id = packageId
                syncTableCursor.user_pin = handler.pin
                syncTableCursor.sync_status = 0
                if syncTableCursor.tryInsert():
                    data["records"]["rec"].append(rec_dict)
                else:
                    if syncTableCursor.tryGet(keyValueList):
                        if syncTableCursor.sync_status != 1:
                            syncTableCursor.date_modify = datetime.datetime.now()
                            syncTableCursor.package_id = packageId
                            syncTableCursor.update()
                            data["records"]["rec"].append(rec_dict)
            else:
                data["records"]["rec"].append(rec_dict)
    context.commit()
    xmlOut = XMLJSONConverter.jsonToXml(json.dumps(data))
    ConnectionPool.putBack(conn)
    try:
        currentTable.close()
        syncTableCursor.close()
    except:
        pass
    logFile = open(mypath + "/output.xml", "w")
    logFile.write('Complete without errors')
    logFile.close()
    return xmlOut

class parserSAXHandlerSyncReport(ContentHandler):
    def __init__(self):
        self.pin = None
        self.grain = None
        self.table = None
        self.param = None
        self.sync = 0
        self.systemId = None
        self.packageId = None
        self.dataWriteFlag = False
        self.commandFlag = False
        self.error = False
        self.errorMes = False
        self.currentData = None

    def startElement(self, name, attrs):
        if name == "command":
            self.commandFlag = True
            for attr, attrValue in attrs.items():
                if attr == "param":
                    self.param = attrValue
                elif attr == "pin":
                    self.pin = attrValue
                elif attr == "grain":
                    self.grain = attrValue
                elif attr == "table":
                    self.table = attrValue
                elif attr == "sync":
                    self.sync = attrValue
                elif attr == "system_id":
                    self.systemId = attrValue
                elif attr == "package_id":
                    self.packageId = attrValue
        if name in ("error", "error_mes"):
            self.dataWriteFlag = True
        if not (self.pin or self.grain or self.table) and name in ("pin", "grain", "table"):
            self.dataWriteFlag = True
        if name == "column":
            self.columnFlag = True

    def characters(self, data):
        if self.dataWriteFlag == True:
            self.currentData = data

    def endElement(self, name):
        if name == "pin" and self.dataWriteFlag == True:
            self.pin = self.currentData
            self.dataWriteFlag = False
        elif name == "grain" and self.dataWriteFlag == True:
            self.grain = self.currentData
            self.dataWriteFlag = False
        elif name == "table" and self.dataWriteFlag == True:
            self.table = self.currentData
            self.dataWriteFlag = False
        if name == "error":
            self.error = self.currentData
            self.dataWriteFlag = False
        if name == "error_mes":
            self.errorMes = self.currentData
            self.dataWriteFlag = False

def syncReport(self, xmlIn):
    parser = make_parser()
    handler = parserSAXHandlerSyncReport()
    parser.setContentHandler(handler)
    parser.parse(StringIO.StringIO(xmlIn.encode('utf8')))
    conn = ConnectionPool.get()
    sesContext = SessionContext('master', 'master')
    context = CallContext(conn, sesContext)
    webServiceUsers = tableCursorImport('webservices', 'web_service_users')(context)
    webServiceUsers.get(handler.pin)
    username = getattr(webServiceUsers, 'login')
    sesContext = SessionContext(username, username)
    userContext = CallContext(conn, sesContext)
    currentTableCursor = tableCursorImport(handler.grain, handler.table)(userContext)
    currentTableMeta = currentTableCursor.meta()
    tableJson = json.loads(currentTableMeta.getCelestaDoc())
    syncGrain, syncTable = tableJson["sync_table"].split('.')
    syncTableCursor = tableCursorImport(syncGrain, syncTable)(context)
    syncTableCursor.setRange("package_id", handler.packageId)
    if handler.error.lower() == 'true':
        for rec in syncTableCursor.iterate():
                rec.error = handler.errorMes
                rec.update()
                userContext.commit()
    else:
        for rec in syncTableCursor.iterate():
                rec.sync_status = 1
                rec.error = None
                rec.update()
                userContext.commit()

    ConnectionPool.putBack(conn)
    try:
        syncTableCursor.close()
        currentTableCursor.close()
    except:
        pass
    return ''

class parserSAXHandlerInsert(ContentHandler):
    def __init__(self):
        self.recFlag = False
        self.dataWriteFlag = False
        self.recordsFlag = False
        self.columnFlag = False
        self.currentNode = None
        self.currentData = None
        self.param = None
        self.pin = None
        self.grain = None
        self.table = None
        self.cursor = None
        self.primaryKey = None
        self.dataToInsert = {"rec":[]}
        self.recordData = {}
        self.gotChars = False
        self.tryUpdate = False
        self.countInserted = 0
        self.ir = 1
        self.systemId = None
    def startElement(self, name, attrs):
        if name == "command":
            self.commandFlag = True
            for attr, attrValue in attrs.items():
                if attr == "param":
                    self.param = attrValue
                    if attrValue == "tryUpdate":
                        self.tryUpdate = True
                elif attr == "pin":
                    self.pin = attrValue
                elif attr == "grain":
                    self.grain = attrValue
                elif attr == "table":
                    self.table = attrValue
                elif attr == "sync":
                    self.sync = attrValue
                elif attr == "system_id":
                    self.systemId = attrValue
        if self.pin is None and name == "pin":
            self.dataWriteFlag = True
        if self.grain is None and name == "grain":
            self.dataWriteFlag = True
        if self.table is None and name == "table":
            self.dataWriteFlag = True
        elif name == "records":
            self.cursor = getTableCursor(self.pin, self.grain, self.table)
            self.primaryKey = self.cursor.meta().getPrimaryKey()
            self.recordsFlag = True
        elif name == "rec":
            self.recFlag = True
        elif self.recFlag == True and name != "rec":
            self.dataWriteFlag = True
            self.columnFlag = True
            self.currentNode = name

    def characters(self, data):
        if self.dataWriteFlag == True:
            self.gotChars = True
            self.currentData = data

    def endElement(self, name):
        if name == "pin" and self.dataWriteFlag == True:
            self.pin = self.currentData
            self.dataWriteFlag = False
        elif name == "grain" and self.dataWriteFlag == True:
            self.grain = self.currentData
            self.dataWriteFlag = False
        elif name == "table" and self.dataWriteFlag == True:
            self.table = self.currentData
            self.dataWriteFlag = False
        elif name == "rec":
            keyValue = []
            if self.tryUpdate:
                for key in self.primaryKey:
                    columnType = self.cursor.meta().getColumn(key).getCelestaType()
                    if columnType == 'INT':
                        keyValueByType = int(self.recordData[key])
                    elif columnType == 'DATETIME':
                        keyValueByType = datetime.datetime.strptime(str(self.recordData[key]).replace("-", ""), '%Y%m%d')
                    elif columnType == 'BIT':
                        if str(self.recordData[key]).lower() == 'false':
                            keyValueByType = False
                        else:
                            keyValueByType = True
                    else:
                        keyValueByType = self.recordData[key]
                    keyValue.append(keyValueByType)
                if self.cursor.tryGet(keyValue):
                    tryGet = True
                else:
                    tryGet = False
            else:
                tryGet = False
            for column, value in self.recordData.items():
                columnType = self.cursor.meta().getColumn(column).getCelestaType()
                if columnType == 'DATETIME' and value is not None:
                    try:
                        valueFormated = value.replace(" ", "").replace("-", "")[0:8]
                        if valueFormated != '':
                            setattr(self.cursor, column, datetime.datetime.strptime(valueFormated, '%Y%m%d'))
                    except:
                        pass
                elif columnType == 'BIT':
                    if str(value).lower() == 'false':
                        setattr(self.cursor, column, False)
                    else:
                        setattr(self.cursor, column, True)
                elif columnType == 'BLOB':
                    getattr(self.cursor, "calc%s" % column)()
                    blobField = self.cursor.__dict__[column].getOutStream()
                    blobField.write(base64.b64decode(value))
                else:
                    setattr(self.cursor, column, value)
            self.cursor.system_id = self.systemId
            if tryGet == True:
                self.cursor.update()
                self.cursor.clear()
            else:
                self.cursor.insert()
                self.cursor.clear()
            self.countInserted += 1
            self.recFlag = False
            self.recordData = {}
        elif self.columnFlag == True and self.recFlag == True and name != "rec" and self.gotChars == True:
            self.recordData[self.currentNode] = self.currentData
            self.dataWriteFlag = False
            self.gotChars = False
        elif name == "records":
            self.cursor.callContext().commit()
            try:
                self.cursor.close()
            except:
                pass
            self.recordsFlag = False

def getTableCursor(pin, grain, table):
    conn = ConnectionPool.get()
    sesContext = SessionContext('master', 'master')
    context = CallContext(conn, sesContext)
    webServiceUsers = tableCursorImport('webservices', 'web_service_users')(context)
    webServiceUsers.get(pin)
    username = getattr(webServiceUsers, 'login')
    sesContext = SessionContext(username, username)
    userContext = CallContext(conn, sesContext)
    currentTable = tableCursorImport(grain, table)(userContext)
    ConnectionPool.putBack(conn)
    try:
        webServiceUsers.close()
    except:
        pass
    return currentTable

def getTableCursorByMaster(grain, table):
    conn = ConnectionPool.get()
    sesContext = SessionContext('master', 'master')
    context = CallContext(conn, sesContext)
    currentTable = tableCursorImport(grain, table)(context)
    ConnectionPool.putBack(conn)
    return currentTable

def getContextByMaster():
    conn = ConnectionPool.get()
    sesContext = SessionContext('master', 'master')
    context = CallContext(conn, sesContext)
    ConnectionPool.putBack(conn)
    return context

def insert(self, xmlIn):
    if not os.path.isdir(r'C:\accred_service_logs'):
        os.makedirs(r'C:\accred_service_logs')
    dateTime = str(datetime.datetime.now().strftime("%Y-%m-%d %H.%M.%S"))
    mypath = r'C:\accred_service_logs/' + "insert_" + dateTime
    if not os.path.isdir(mypath):
        os.makedirs(mypath)
    logFile = open(mypath + "/input.xml", "w")
    logFile.write(xmlIn)
    logFile.close()
    parser = make_parser()
    handler = parserSAXHandlerInsert()
    parser.setContentHandler(handler)
    parser.parse(StringIO.StringIO(xmlIn.encode('utf8')))
    data = {"result":"(" + str(handler.countInserted) + " row(s) affected)"}
    xmlOut = XMLJSONConverter.jsonToXml(json.dumps(data))
    logFile = open(mypath + "/output.xml", "w")
    logFile.write(xmlOut)
    logFile.close()
    return xmlOut

def insertFromFile(self, fileName):
    parser = make_parser()
    handler = parserSAXHandlerInsert()
    parser.setContentHandler(handler)
    parser.parse(fileName)
    data = {"result":"(" + str(handler.countInserted) + " row(s) affected)"}
    xmlOut = XMLJSONConverter.jsonToXml(json.dumps(data))
    return xmlOut

def dataBaseXMLUpload(self, xmlIn):
    u'''Загрузка данных в таблицу через веб-сервис.
     Данные получены в стандартном XML из common.dbutils '''
    
    if not os.path.isdir(r'C:\accred_service_logs'):
        os.makedirs(r'C:\accred_service_logs')
    dateTime = str(datetime.datetime.now().strftime("%Y-%m-%d %H.%M.%S"))
    mypath = r'C:\accred_service_logs/'+"dbupload_"+dateTime
    if not os.path.isdir(mypath):
        os.makedirs(mypath)
    logFile = open(mypath+"/input.xml", "w")
    logFile.write(xmlIn)
    logFile.close()
    # работа с XML
    req = re.sub(r'\<show\:requestAnyXML\>|\<\/show\:requestAnyXML\>', '', xmlIn, flags=re.U)
    root = ET.fromstring(req)
    command_json = root.attrib # атрибуты команды
    # ищем таблицу и преобразуем ее содержимое в строку
    data = root.find('data').find('table')
    xml_data = ET.tostring(data)
    tableInstance = getTableCursor(command_json['pin'], command_json['grain'], command_json['table'])
    dataStream = ByteArrayInputStream(String(xml_data).getBytes("UTF-8"))
    exchange = DataBaseXMLExchange(dataStream, tableInstance)
    exchange.uploadXML(command_json['param'] or 'ui')
    dataStream.close()
    # вывод результата  
    data = {"result":"Ok"}
    xmlOut = XMLJSONConverter.jsonToXml(json.dumps(data))
    logFile = open(mypath+"/output.xml", "w")
    logFile.write(xmlOut)
    logFile.close()
    return xmlOut

def uploadCursor(context, cursor_obj, grain_name, cursor_name, systemName, userPin, wsUrl, filteredCursor=True, dbaction='ui'):
    u'''Выполняет выгрузку cursor через веб-сервис.
    filteredCursor - если стоит True, значит получаем данные через DataBaseXMLExchange.downloadXML() - курсор
    должен быть отфильтрован, иначе формируем XML по данным, которые сейчас есть в курсоре.'''

    wsClient = Client(wsUrl)
    
    # выгружаем выбранные записи в текст
    if filteredCursor:
        baos = ByteArrayOutputStream()
        exchange = DataBaseXMLExchange(baos, cursor_obj)
        exchange.downloadXML()
        data = baos.toString("UTF-8")
        baos.close()
    else:
        data_dict = {'table': {'@name': cursor_name,
                          'row': {'field': []}}
                }
        # текущая запись
        for field in cursor_obj.meta().getColumns():
            if cursor_obj.meta().columns[field].getCelestaType() == 'BLOB':
                # пока не будем реализовывать
                pass
            else:
                fd = {}
                fd['@name'] = field
                fd['#text'] = unicode(getattr(cursor_obj, field)) if getattr(cursor_obj, field) is not None and getattr(cursor_obj, field)!='' else u' None '
                data_dict['table']['row']['field'].append(fd)
        data = XMLJSONConverter.jsonToXml(json.dumps(data_dict))

    wsRequest = u'''
        <show:requestAnyXML>
            <command 
            param="%s" 
            system_id="%s" 
            pin="%s" 
            grain="%s" 
            table="%s">
                <data>
                    %s
                </data>
            </command>
        </show:requestAnyXML>
    ''' % (dbaction, systemName, userPin,
           grain_name,
           cursor_name,
            data.replace('<?xml version="1.0" encoding="UTF-8"?>', '').strip())
    wsXmlOut = wsClient.service.handle(request=wsRequest, procName="webservices.functions.dataBaseXMLUpload.celesta")
    if '<error' not in wsXmlOut:
        pass
    else:
        context.message(u'При выгрузке возникли ошибки.')
        
    cursor_obj.close()